;;; eggdoc-texinfo
;; Copyright (c) 2006 Jim Ursetto
;; $Revision$

;; To do:
;; DONE Function collapsing (works great)
;; DONE Fixed extra whitespace at beginning of line in eggdoc-objc -- info does not elide it.
;; DONE (author) etc tags should really be parsed--a (url) tag within this
;;   will be runtogether -- see sqlite3
;; DONE Rudimentary table tag parsing  -- see sqlite3
;; SKIP symbol-table -- should we use @vtable?
;; DONE Read-syntax special form, and safe-read
;; DONE Index is generated (but variables containing : cannot serve as valid index nodes)
;; SKIP Handle quotes (but q tag works)
;; DONE Fixed handling of custom stylesheets -- make sure they operate on partitioned doc.
;; SKIP Detailed node listing
;; DONE Multitable uses max-width of .30 for first column for aesthetic reasons
;;      and only 90% width total to lessen accidental hbox runovers
;; DONE : Fixed verbatim formatting in hostinfo, doctype
;; DONE : Fixed incorrect section parens in args.egg, amb
;; DONE : Fixed excess p whitespace (info does not collapse) in args.egg
;; DONE : fixed invalid read (typo) in ajax
;; DONE : strip commas from nodenames.  Tried to use @comma{}, but a bug in texi2pdf
;;        prevents it.
;; License section of sqlite3 doc.scm is under documentation section and won't parse.
;;    This is because license is hardcoded to 'section'.  commit new doc
;; Info directory entry for egg
;;    -- Install with:
;;       install -m 644 hostinfo.info /usr/local/info
;;       install-info --info-dir=/usr/local/info hostinfo.info
;; -- If your custom stylesheet is not confined to macros, and generates html elements,
;;    it will not work unchanged.
;; -- Caveats
;; emph across paragraphs doesn't work.  Use blockquote or emph each para.
;; Reformatting pre tags may be required to fit the page.
;; Info may not delete some whitespace between words.  This typically occurs
;;    when you have a single string that spans multiple lines.  Recommended
;;    solution is to close the string tag 
;; There should be no bare strings in section bodies: they must be wrapped in block
;;   tags--PRE or P, for example.
;;   (section "License" "Public domain") is invalid, but happens to work in HTML.
;;   (section "License" (p "Public domain")) is correct.
;; Texinfo may not like certain non-ASCII characters -- see fp.egg.
;; 

(require-extension sxml-transforms)
(require-extension sxml-tools)
(require-extension regex)

(declare
 (export eggdoc:make-texinfo-stylesheet
         eggdoc->texinfo
         eggdoc->texinfo-sxml
         texinfo-sxml->texinfo
         universal-texi-conversion-rules
         eggdoc:make-defsig
         ))

;; export: eggdoc:make-defsig

;; special chars: @, {, }
;; Begin/end quotes are not translated.
(define string->goodTexi
  (make-char-quotator '((#\{ . "@{") (#\} . "@}") (#\@ . "@@"))))

;; Procedure to grab contents of context node (not including args).
(define elt-contents (sxpath '((*not* args))))
;; Grab arguments to this node--flattens multple occurrences of (args ...) into one list.
(define elt-args
  (sxpath '(args *any*)))
;; Grab element by name given context node.
(define-macro (elt name context)
  `((sxpath (list ,name)) ,context))
(define (elt-text name context)
  ((sxpath (list name '// '*text*)) context))

;; Join args together based on tag.  Some args are joined with commas,
;; some with spaces, some with @|, some surrounded by braces.
;; Consistency is not the strong suit here.  strip-commas is used
;; to compensate for a bug in texi2pdf where the tex indexer barfs
;; on @comma{} in a nodename.  Because it is likely to barf on
;; other info commands as well, it will probably be necessary to
;; create a separate (node ...) transformer which strips out or flattens
;; tags as in direntry-node.
;; Node text is not transformed in the document itself--so we must
;; also do the transform in node-titles, so make-menu works.
;; This really annoys me.
;; Comma-strip functionality might better go in bare-title, as this
;; is called for every node title created from a section anyway.
(define (strip-commas args)
  (let ((quotator (make-char-quotator '((#\, . "")))))
    (post-order args `((*text* . ,(lambda (trigger str)
                                    (quotator str)))
                       (*default* . ,list)))))
(define (quote-commas args)
  (let ((quotator (make-char-quotator '((#\, . "@comma{}")))))
    (post-order args `((*text* . ,(lambda (trigger str)
                                        (quotator str)))
                       (*default* . ,list)))))
(define (args-join tag args)
  (cond ((memq tag '(multitable syncodeindex))
         (intersperse args " "))
        ((memq tag '(node))
         (intersperse (strip-commas args) ", "))
        (else
         (intersperse (quote-commas args) ", "))))

(define (eat-tag . expr)
  (if (pair? expr)
      (cdr expr)
      '() ))
(define (discard . args) '())
(define (discard-and-warn . args)
  (warning (conc "eggdoc-texinfo: discarded " args))
  '())

;; Several different types of commands:
;; Chapter structuring commands (@chapter xxx, @section xxx, @top xxx) -- no end tag.
;;     In Info, these just generate headings with underlining--no actual node linkage.
;; Node structuring commands (@node xxx) -- valid until next occurence of @node
;; menu (@menu / @end menu)
;; Other wrappers (@tag / @end tag) --@verbatim, @itemize  -- block-enclosing
;; Single item tags (@item xxx)
;; Inline tags (@tag{xxx, ...}) such as @emph{...} and @xref{nodename, crossrefname}.
;; Note: Although most wrapper tags don't have an argument--the text in the body
;;       is the argument--some do, such as @table @... \n body \n @end table
;;       or @quotation <optional-arg> \n body \n @end quotation.
;; These tags must be encoded separately
;; Looks like the block tags with multiple (more than one) argument
;; separated by spaces are @def*, @macro, @syn[code]index, @multitable.
;; @node has up to 4 args separated by commas.
;; @oddfooting/@oddheading are separated by @|.
;; @float: two optional args, separated by commas.
;; Some tags (@def*, @multitable) accept brace-surrounded arguments, if the
;; argument contains spaces.

;; @anchor / @xref for in-document references
;; @url / @uref for url references (sec 8.9; compare to @indicateurl)

;; Note: verb / verbatim should not evaluate their arguments nor use string->goodTexi.
;; verb needs to be handled specially.
;; @itemize @bullet{}   <-- braces are legal

;; makeinfo --xml: each tag argument has its own tag, often prefixed with the master tag name.
;; e.g. itemize -> itemfunction; definition -> defcategory, defvariable, deffunction;
;; uref -> urlref, urldesc; node -> nodename, nodenext, nodeup, nodeprev;
;; chapter -> title; section -> title (interestingly the same name...)
;; Paragraphs wrapped in <para>.  Chapter structuring commands are paired, not single
;; tags (<chapter><title>T</title>...</chapter>, not <chapter title="T" />...).
;; The official TexinfoML DTD is at <http://www.gnu.org/software/texinfo/dtd/4.7/texinfo.dtd>.


(define flatten-url-rule
  `(url ((args . ,(lambda (tag href . desc)
                    (if (pair? desc) desc href))))
        . ,(lambda (tag . body) body)))

(define universal-texi-conversion-rules
  `((args . ,list)
    (texinfo:begin . ,eat-tag)

    ;; So I can push text literals in emergencies -- for example, the command in @multitable. 
    (literal ((*text* . ,(lambda (trigger str) str)))
             . ,(lambda (tag . body) body))
    
    (*text* . ,(lambda (trigger str) (if (string? str)
                                    (string->goodTexi str)
                                    str)))
    
    (*default* .
               ,(let* ((block-tags '(itemize enumerate verbatim format display menu ifnottex
                                     quotation table ftable vtable lisp multitable
                                     titlepage direntry))
                       (block-tag? (lambda (t) (memq t block-tags)))
                       (brace-tags '(
                                     ;; 1+ arg
                                     acronym email image inforef pxref ref uref url xref
                                     ;; 1 arg
                                     code kbd key samp verb var env file
                                     command option dfn cite indicateurl
                                     emph strong sc b i t r footnote 
                                     ;; 0-arg
                                     bullet comma copyright dots enddots equiv error
                                     expansion TeX LaTeX minus point print result
                                     today))
                       (brace-tag? (lambda (t) (memq t brace-tags))))
                 (lambda (tag . elts)
                   (let ((expr (cons tag elts)))    ;; for sxpath
                     ; (print expr)
                     (if (brace-tag? tag)
                         (list #\@ tag #\{
                               (intersperse (elt-args expr) ", ")
                               (elt-contents expr)
                               #\})
                         (list #\newline
                               (and (block-tag? tag) #\newline)
                               #\@ tag
                               (list #\space
                                     (args-join tag (elt-args expr))
                                     (if (block-tag? tag) #\newline #f))
                               (elt-contents expr) #\newline
                               (and (block-tag? tag)
                                    (list "@end " tag))))))))
    ;; p tag -- just group a paragraph.  Right now only here to provide
    ;; whitespace around the paragraph.
    (p . ,(lambda (tag . elts)
            (list #\newline #\newline elts)))
    (menu-node . ,(lambda (tag . body)
                    (list #\* #\space body #\: #\: #\newline)))
    (texinfo-header . ,(lambda (tag)
                         (list #\\ "input texinfo @c -*- texinfo -*-" #\newline)))

    ;; Treat verbatim text separately.  We render the tag manually; if we want
    ;; to go through normal universal-conversion rules, we must remove this
    ;; tag from the universal-conversion rules and do a post-order on universal-conversion.
    ;; Note: tried adding extra preceding newline here to compensate for missing
    ;; space, and failed (extra space always added).  See 1.5 Classes
    ;; Note: verbatim doesn't handle tags, so we make sure to flatten url.
    ;;       If tag parsing becomes necessary, consider using @example for pre.
    (verbatim ( ,flatten-url-rule
               (verbatim . ,eat-tag)
               (*text* . ,(lambda (trigger str) str)))
              . ,(lambda (tag . body) `(#\newline #\newline #\@ "verbatim" #\newline
                                   ,@body #\newline
                                   #\@ "end verbatim" #\newline)))

    (deffn . ,(lambda (tag category name args . body)
                (list #\newline #\@ "deffn " #\{ category #\} #\space
                      #\{ name #\} (intersperse args " ") #\newline
                      body #\newline
                      #\@ "end deffn" #\newline)))
    (deffnx . ,(lambda (tag category name)
                 (list #\newline #\@ "deffnx " #\{ category #\} #\space
                       #\{ name #\})))


    ;; This is a kludge to remove extraneous whitespace within direntry.
    (direntry . ,(lambda (tag . nodes)
                   (list #\newline
                         #\@ tag #\newline
                         nodes
                         #\@ "end " tag)))
    ;; Create a direntry node; this collapses any URLs.
    (direntry-node ( ,flatten-url-rule )
                   . ,(lambda (tag entry-name filename desc)
                        (list "* " entry-name ": (" filename ")." #\tab #\tab desc #\newline)))

    (br . ,(lambda (tag) #\newline))   ;; Maybe this should be "@\n".
    (q  . ,(lambda (tag . body)
             (list #\` #\` body #\' #\')))

    ;; Comma-separated args for brace tags.  For single-arg tags the
    ;; bare value can be used, without this tag.
    ;; (emph (args "please" (strong "please") "me")) -> @emph{please, @strong{please}, me}
    ;; (emph "oh yeah")                              -> @emph{oh yeah}
    ;; (emph (args ("like" "I") ("please" "you")))   -> @emph{likeI, pleaseyou}

    ;; Removed due to conflict with block args tag
;;     (args . ,(lambda (tag . args)
;;                (intersperse args ", ")))

    ))

(define text-only (sxpath '(// *text*)))
(define (bare-title x)
  (if (pair? x)
      (if (string? (car x))
          x
          (text-only x))
      x))

;; Scrape node titles: used on texinfo-sxml output, such as will be passed
;; in as the body parameter to a section tag in the Eggdoc SXML stylesheet.
;; Handles non-arg and args version (kind of hacky).  We must strip commas
;; here to match the output produced by the node transformer (unfortunately).
(define (node-titles body)
  (map strip-commas
       (map (lambda (n)
              (let ((arg (second n)))
                (if (and (pair? arg)
                         (eq? 'args (car arg)))
                    (second arg)
                    arg)))
            ((sxpath '(node)) (cons '*dummy* body)))))

;; Used so we can insert a menu between nodes and non-nodes (all non-nodes
;; will be gathered together at the beginning; however, they should always
;; be there anyway.)

(define (select-nodes body)
  ((sxpath '((node))) (cons '*dummy* body)))
(define (select-non-nodes body)
  ((sxpath '((*not* node))) (cons '*dummy* body)))

(define (latest-version doc)
  (let ((path ((sxpath '(eggdoc:about ((history (version 1)))))
               doc)))
    (if (null? path)
        #f
        (cadar path))))

;; Make node menu, using node titles from the nodes in body.
(define (make-menu body)
  (let ((child-nodes (map (lambda (x)
                            (list 'menu-node x))
                          (node-titles body))))
     (if (null? child-nodes)
         #f
         `(menu ,@child-nodes))))

;; Return a function suitable for generating a @deffn tag in CATEGORY.
(define (car-if-pair x)
  (if (pair? x) (car x) x))

;; Not used (obsoleted by make-defsig and make-deffn-multi) but here to show the basic
;; structure of make-deffn-multi.
(define (make-deffn category)
  (lambda (tag sig . body)
    (let ((title (symbol->string (car-if-pair (with-input-from-string sig read)))))
      `(deffn ,category ,title
         ()
         (lisp ,(if (equal? title sig) #f sig)) ;; Print entire signature if differs from title.
                                                ;; Mostly useful for "record" types.
                                                ;; Probably record should be @defvr anyway.
         ,@body))))

;; Helper function -- convert e.g. a (parameter) tag into a definition/signatures group.
(define eggdoc:make-defsig
  (lambda (tag sig . body)
    `(definition (signatures (signature ,(symbol->string tag) ,sig))
       ,body)))
(define make-defsig eggdoc:make-defsig)
(define (make-section type)
  (lambda (tag title . body)
    `(node (args ,(bare-title title))
           (,type ,title
                  ,(select-non-nodes body)
                  ,(make-menu body)
                  ,(select-nodes body)))))

(define (read-from-string str)
  (condition-case (with-input-from-string str read)
                  [(exn)
                   (warning (conc "read-from-string: invalid read: " str))
                   str]))

;; Reading the signature with (read) will not work for read-syntax.  We now
;; have a read-syntax tag which does not parse the signature string.  However,
;; I will leave this logic, which checks if the category tag begins with
;; "read", in for now.  This is mainly for use in (definition ... ) lists.
(define (read-syntax? str)
  (string-search "^read" str))   ;; reader macro, read-syntax, read syntax, ...

(define (make-deffn-multi categories)
  (lambda (tag sigs . body)
    (let ((titles  (map (lambda (s c)
                          (if (read-syntax? c)  ;; Could be removed.
                              s
                              (->string (car-if-pair (read-from-string s)))))
                        sigs categories)))
      (receive (collapsed-categories collapsed-titles)
          (collapse-two categories titles)
        `(deffn ,(car categories) ,(car titles)
           () ;; always empty args
           ,(map (lambda (cat tit)
                   `(deffnx ,cat ,tit))
                 (cdr collapsed-categories)
                 (cdr collapsed-titles)) ;; Collapse equal adjacent titles.
           ;; No line-wrapping done, which can lead to hbox overflows on TeX output.
           ,(let ((sig-lines (map (lambda (title sig)
                                      (if (equal? title sig)
                                          #f
                                          (list sig)))  ;; Use entire sig only if differs from title. 
                                    titles sigs)))
              (and (any identity sig-lines) ;; non-empty
                  `(lisp ,(intersperse sig-lines     
                                       #\newline)))) ;; newline is horrible hack.
           ;; Mostly useful for "record" types.
           ;; Probably record should be @defvr anyway.
           (p ,@body))))))


;; Discard internal links.
(define (ensure-external-href x)
  (cond ((not (string? x))
         #f)
        ((string-search "^(http:|mailto:)" x)
         x)
        (else #f)))


;; Returns a list such that adjacent eq? elements in L are collapsed
;; down into a single element.
(define (collapse L)
  (fold-right (lambda (x xs)
                (if (and (pair? xs)
                         (equal? x (car xs)))
                    xs
                    (cons x xs)))
              '()
              L))
;; Zip L and M into one list and collapse it, then unzip and return the two lists.
(define (collapse-two L M)
  (let ((R (collapse (zip L M))))
    (values (map first R)
            (map second R))))

;; Not currently used

;; Pull out certain headers and wrap them in an eggdoc:about tag.  Creating
;; a separate node for these headers is too difficult or messy using only
;; stylesheet transforms or sxpath on toplevel tags.
(define (partition-about doc)
  (receive (headers body)
      (partition (lambda (x)
                   (memq (car x)
                         '(requires history usage)))
                 (cdar doc))
    `((eggdoc:begin
       (eggdoc:about ,@headers)
       ,@body))))


(define (eggdoc:make-texinfo-stylesheet unpartitioned-doc)
  (let* ((doc (partition-about unpartitioned-doc))
         (basic-conversion-rules
          `(
            (*text* . ,(lambda (trigger str) str))

            (ul . ,(lambda (tag . elts)
                     `(itemize ,@elts)))
            (ol . ,(lambda (tag . elts)
                     `(enumerate ,@elts)))
            (li . ,(lambda (tag . elts)
                     `(item ,@elts)))
            (em . ,(lambda (tag . elts)
                     `(emph ,@elts)))
            (strong . ,(lambda (tag . elts)
                         `(strong ,@elts)))
            (b . ,(lambda (tag . elts)
                    `(strong ,@elts)))
            (i . ,(lambda (tag . elts)
                    `(emph ,@elts)))
            (tt . ,(lambda (tag . elts)
                     `(code ,@elts)))
            (code . ,(lambda (tag . elts)
                       `(code ,@elts)))
            (& . ,(lambda (tag . elts)
                    (map (lambda (e)
                           (switch (string->symbol e)
                                   ('copyright '(copyright))
                                   ('lambda "lambda")
                                   ('mdash "---")
                                   ('ndash "--")
                                   (else (warning (conc "unhandled & element: " e)))))
                         elts)))
            (p    . ,list)
            (div  . ,eat-tag)
            (span . ,eat-tag)
            (a ((@ ((href . ,(lambda (tag value)
                               (and-let* ((url (ensure-external-href value)))
                                 `(href ,url)))) ;; Allow only href http: or mailto:.
                    (*default* . ,(constantly '())))
                   . ,(lambda (trigger . value)
                        (cons '@ value))))
               . ,(lambda elems
                    ;; We use sxpath to pull out href elements, because we transformed non-href
                    ;; elements into empty elements (rather than eliminating them).
                    (let ((link ((sxpath '(@ href *text*)) elems))
                          (contents ((sxpath '((*not* @))) elems))) ;; get all text and elts (non-@)
                      (if (null? link) ;; Not an href= -- a name=, for example
                          contents
                          (if (pair? contents)
                              `(url (args ,link ,contents))
                              `(url ,link))))))

            (url . ,(lambda (tag link . contents)
                      (let ((link (ensure-external-href link)))
                        (if link
                            (if (pair? contents)
                                `(url (args ,link ,contents)) ;; Note that contents is not spliced in
                                `(url ,link)) ;; to ensure it remains one argument.
                            ;; Internal link; use plain contents.
                            contents))))))
         ;; Run post-order on tag contents and return contents.  Also eats
         ;; p tags as they don't make sense in an author or description context
         ;; (we should actually eat all block tags).
             (transform-elt (lambda (tag) (post-order (cdar (elt tag doc))
                                                 `((p . ,eat-tag)
                                                   ,@basic-conversion-rules))))
             (egg-name (elt-text 'name doc))) ;; here so Usage can access it

;; Begin stylesheet
    `(
      ,@basic-conversion-rules
      (eggdoc:begin
       ((name        . ,discard) ;; These still act recursively, so just as well could be
        (description . ,discard) ;; at toplevel.  These are discarded because their transforms
        (author      . ,discard) ;; are unused--instead sxpath is used on the document.
;;      (history     . ,discard)
;;      (version     . ,discard)
;;      (requires    . ,discard)
;;      (usage       . ,discard)
        (download    . ,discard))
       . ,(lambda (tag . elts)
            (let ((egg-desc (transform-elt 'description))
                  (egg-ver  (latest-version doc))
                  (egg-author (transform-elt 'author)))
              `(texinfo:begin
                (texinfo-header)
                (c "Generated by eggdoc-texinfo $Revision$ (c) 2006 Jim Ursetto")
                (c "%**start of header")
                (setfilename ,egg-name ".egg.info")
                (settitle ,egg-name " egg: " ,egg-desc)
                (c "%**end of header")
                (finalout)  ;; Don't print black overfull hboxes
                (dircategory "The Algorithmic Language Scheme"
                             (direntry
                              (direntry-node (,egg-name ".egg")
                                             (,egg-name ".egg")
                                             ,egg-desc)))
                
                (titlepage
                 (title ,egg-name " egg")
                 (subtitle ,egg-desc)
                 (subtitle "Extension for Chicken Scheme")
                 ,(and egg-ver
                       `(subtitle "Version " ,egg-ver))
                 (author ,egg-author))
                (contents)
                (syncodeindex (args "vr" "fn"))
                (ifnottex
                 (node (args "Top")
                       (top ,egg-name " egg")
                       (p ,egg-desc)
                       (p (noindent) "Written by " ,egg-author)
                       ,(and egg-ver
                             `(p "This manual corresponds to version " ,egg-ver
                                 " of the " ,egg-name " extension library for Chicken Scheme."))
                       ;; Version history node
                       ;; Requires

                       ,(make-menu
                         `( ,@elts
                            (node (args "Index"))))))
                
                ,@elts
                (node (args "Index")
                      (unnumbered Index)
                      (printindex fn))

                (bye)))))

    
      (section . ,(make-section 'chapter))
      (subsection . ,(make-section 'section))
      (subsubsection . ,(make-section 'subsection))
      (subsubsubsection . ,(make-section 'subsubsection))
      
      
      (documentation *macro* . ,(lambda (tag . body) `(section "Documentation" ,@body)))
      (license       *macro* . ,(lambda (tag . body) `(section "License" (pre ,@body))))
      (examples      *macro* . ,(lambda (tag . body) `(section "Examples" ,@body)))

;;; About
;; Certain elements are wrapped (externally) into an eggdoc:about tag, which
;; we convert into a section of its own.
      (eggdoc:about  *macro* . ,(lambda (tag . body)
                                  `(section "About this egg"
                                            ,@body)))
      (history       *macro* . ,(lambda (tag . body)
                                  `(subsection "Version history"
                                     (version-table ,@body))))
      (version       *macro* . ,(lambda (tag ver . body)
                                  `((describe ,ver ,body))))
      (usage         *macro* . ,(lambda (tag . body)
                                  `(subsection "Usage"
                                     (p "Load this egg like so: ")
                                     (p 
                                      ,(if (null? body)
                                           `(tt "(require-extension " ,egg-name ")")
                                           body)))))
      (requires      *macro* .  ,(lambda (tag . requires)
                                   `(subsection "Requirements"
                                     (p "This egg requires the following extensions:")
                                     (p ,(intersperse (map (lambda (x) `(tt ,x))
                                                           requires)
                                          ", ")))))


      
      (pre . ,(lambda (tag . body)
            `(verbatim ,@body)))
      (group . ,(lambda (tag . body) body)) ;; Temporary


      (procedure *macro* . ,make-defsig)
      (macro     *macro* . ,make-defsig)
      (record    *macro* . ,make-defsig)
      (read-syntax . ,(lambda (tag sig . body)
                        `(deffn "read-syntax" ,sig ()
                           (p ,body))))
      (parameter *macro* . ,make-defsig)
      ;; Use in place of parameter signature.  Expects bare parameter name
      ;; [not wrapped in parens] and default value, as strings.
      (with-default-param *macro*
                          . ,(lambda (tag param default)
                               (conc "(" param ") [default: " default "]")))
      
      ;; It -seems- that table and pre both need an extra preceding newline (might
      ;; not be true for record immediately followed by table).
      (version-table . ,(lambda (tag . descs)
                          `(table (args @code) ,@descs)))

      (symbol-table . ,(lambda (tag . descs)
                         `(table (args @code) ,@descs)))

      ;; (div) tags inside describe don't work.
      (describe . ,(lambda (tag sym desc)
                     `((item ,sym) ,desc)))

;;; Definitions
      (definition . ,(lambda (tag term . body)
                       (let* ((sig-exprs ((sxpath '(signature)) (cons '*dummy-tag* term)))
                              (categories (map second sig-exprs))
                              (sigs (map third sig-exprs)))
                         ((make-deffn-multi categories) tag sigs body))))
      (signatures . ,eat-tag)
      (signature  . ,list)

;;; tables
      ;; Try to handle table tags in a reasonable way.  Space is split evenly between
      ;; all columns, as Texinfo requires you to calculate width manually or provide
      ;; the longest string of text per column--it doesn't flow automatically.
      (table ((tr . ,list)
              (td . ,list)
              (th . ,list)
              (@ *macro* . ,discard-and-warn))
             . ,(lambda expr
                  (let* ((header-rows  ((sxpath '((tr (th))))    ;; list of (tr (th ...)) sexprs
                                          expr))
                         (header-lists (and (pair? header-rows)  ;; just the data from TH nodes
                                            (map cdr (cdar header-rows))))
                        (data-rows
                         ((sxpath '((tr (td)))) expr))) ;; list of (tr (td ...)) expressions
                    `(multitable (args (literal "@columnfractions")
                                       ,@(let* ((first-data-list (cdar data-rows))
                                                (len (length first-data-list)))
                                           (if (<= len 1)
                                               (list 1.0)   ;; below expr can handle len==1, though
                                               (let ((first-column-width
                                                      (min .30 (/ 1.0 len))))
                                                 (cons first-column-width
                                                  (map (lambda (x)
                                                         (exact->inexact (/ (- 1.0 first-column-width
                                                                               0.1)  ;; fudge for tex bug
                                                                            (- len 1))))
                                                       (cdr first-data-list)))))))
                                 ,(and (pair? header-lists)
                                       `(headitem
                                         ,(car header-lists)
                                         ,(map (lambda (x) (cons 'tab x))
                                               (cdr header-lists))))

                                 ,(and (pair? data-rows)
                                       (map (lambda (row)
                                              `(item
                                                ,(cdr (second row))
                                                ,(map (lambda (x)
                                                        (cons 'tab (cdr x)))
                                                      (cddr row))))
                                            data-rows))))))


      (dl . ,(lambda (tag . body)
                       `(table (args @asis)
                               ,@body)))
      (dt . ,(lambda (tag . body)
                       `(item ,body)))
      (dd . ,(lambda (tag . body)
                       body))
      (br . ,list)
      (q  . ,list)

      (blockquote . ,(lambda (tag . body)
                       `(quotation ,@body)))

      
      (*text* . ,(lambda (trigger str) str))

      (@ *macro* . ,discard-and-warn)
    ; (*default* . ,(lambda elts (list)))     ;; When commmented out, do NOT accept unknown SXML

      )))


(define (eggdoc->texinfo-sxml doc . rest)
  (let-optionals rest ((ss (eggdoc:make-texinfo-stylesheet doc)))
    (pre-post-order (partition-about doc) ss)))

(define (texinfo-sxml->texinfo doc . rest)
  (let-optionals rest ((ss universal-texi-conversion-rules))
    (SRV:send-reply (pre-post-order doc ss))))

;; This optional stylesheet arg modifies the sxml->texinfo stylesheet.
(define (eggdoc->texinfo doc . ss)
  (texinfo-sxml->texinfo
   (if (null? ss)
       (eggdoc->texinfo-sxml doc)
       (eggdoc->texinfo-sxml doc (car ss)))))

;; Debugging

;;; debugging
(define (out)
  (with-output-to-file "/tmp/example.texi"
    (lambda ()
      (eggdoc->texinfo doc))))

